import os
import shutil
import unittest


class TestsExoml(unittest.TestCase):
    def test_setup_files(self):
        self.assertTrue(True)
        #IATSON_planet()


if __name__ == '__main__':
    unittest.main()
